#include "Robot_io.h"
#include <stk600/stk_io.h>
#include <stk600/usart.h>
#include <stk600/timer.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <stdlib.h>
#include <stdio.h>

/**************************************
Communication protocol
=======================

All messages except ACK are terminated with cr (\r)

c represents checksum

Messages to simulator:
-------------------------
?c    		reset (test connection) Besvares med ACK ( \x06)
Oxc             has hit obstacle, x 1 or 0 (true or false)
Responses to requests:
Hxxxc    	heading,  xxx desimal digits,
Pxxxyyyc  	position,  xxx desimal digits, yyy desimal digits,
Dxxxxc     	radar distance xxxx desimal digits,

Messages from simulator:
------------------------
Commands:
rc	  reset
Wpxxc	  swing, p = ' ' or '-', "xx" desimal digits (or space)
Spxxc	  speed, p = ' ' or '-', "xx" desimal digits (or space)
Rxxxc     radar ping  xxx desimal digits (or space), direction
Requests:
Hc       send compass heading
Pc       send position

****************************************/

static void addChecksum(char * dest, char a[], int n);
static boolean checksumOK(char a[], int n);

static boolean running = false;

static unsigned int heading = 0;
static position_type pos = {0,0};
static unsigned int radar_dist = 0;
static boolean hasHitObstacle = false;

static boolean radarDistReady = false;
static boolean positionReady = false;
static boolean headingReady = false;

static Timer T;

static char 	buf[10];

static int		i = 0;

static enum {wait_first_char, receiving_message} state = wait_first_char;

static void init(void)
{
	running = true;
	//initialize timer 1
	cli();
    OCR1A = (F_CPU/8)/100;
    TCCR1B = (1<<CS11); 	// prescaler 8
    TIMSK = (1<<OCIE1A);	//enable OCA interrupt
    TIFR = (1<<OCF1A);  	//clear flaggregister
    TCNT1 = 0;             	//clear counter register
	usartInstall(115200);
	sei();
	i = 0;
	T = newTimer();
	state = wait_first_char;
	char str[4];
	addChecksum(str, "r", 1);
	usartWriteString(str);
}

static void communication_process(void)
{
	char c;
	char str[5];
	while (usartDataIsPresent())
	{

		switch(state)
		{
			case wait_first_char:
				c = usartRead();
				switch (c)
				{
					case '?':
					case 'H':
					case 'P':
					case 'D':
					case 'O':
						buf[0] = c;
						i = 1;
						state = receiving_message;
						break;
					default:
						// ignore
						break;
				}
				break;
			case receiving_message:
				c = usartRead();
				if ((c == '\r') || (i >= 8))
				{
					buf[i] = '\0';
					i = 0;
					//analyse command
					switch(buf[0])
					{
						case '?':
							if (checksumOK(buf,1))
							{
								usartWrite('\x06'); //send ACK
								cli();
								UCSR1B = 0;  // Kill USART1
								asm(" jmp 0"); // RESET!
							}
							break;
						case 'H':
							if (checksumOK(buf,4))
							{
								for (int k = 0; k < 3; k++)
									str[k] = buf[k+1];
								str[3] = '\0';
								heading = atoi(str);
								headingReady = true;
							}
							break;
						case 'P':
							if (checksumOK(buf,7))
							{
								for (int k = 0; k < 3; k++)
									str[k] = buf[k+1];
								str[3] = '\0';
								pos.x = atoi(str);
								for (int k = 0; k < 3; k++)
									str[k] = buf[k+4];
								str[3] = '\0';
								pos.y = atoi(str);
								positionReady = true;
							}
							break;
						case 'D':
							if (checksumOK(buf,5))
							{
								for (int k = 0; k < 4; k++)
									str[k] = buf[k+1];
								str[4] = '\0';
								radar_dist = atoi(str);
								radarDistReady = true;
							}
							break;
						case 'O':
							if (checksumOK(buf,2))
							{
								hasHitObstacle = (buf[1] == '1');
							}
							break;
					}
					state = wait_first_char;
				}
				else
				{
					buf[i++] = c;
				}
				break;
		}
	}
}

ISR(SIG_OUTPUT_COMPARE1A)
{
    //k OCR1A med tikk_tall;
    OCR1A += (F_CPU/8)/100;
	communication_process();
}

void setSwing(int swing)
{
	char str1[7], str2[8];
	if (!running) init();
	if (swing < -10)
		swing = -10;
	else if (swing > 10)
		swing = 10;

	sprintf(str1,"W%03d",swing);
	addChecksum(str2,str1,4);
	usartWriteString(str2);
}

void setSpeed(int speed)
{
	char str1[7], str2[8];
	if (!running) init();
	if (speed < -10)
		speed = -10;
	else if (speed > 10)
		speed = 10;
	sprintf(str1,"S%03d",speed);
	addChecksum(str2,str1,4);
	usartWriteString(str2);
}

void radarPing(int direction)
{
	char str1[7], str2[8];
	if (!running) init();
	if (direction >= 360)
		direction = 359;
	sprintf(str1,"R%03u",direction);
	addChecksum(str2,str1,4);
	usartWriteString(str2);
	radarDistReady = false;
}

boolean radarDataReady(unsigned int * distance)
{
	if (radarDistReady)
	{
		*distance = radar_dist;
		return true;
	}
	return false;
}

void getPosition()
{
	char str[4];
	addChecksum(str, "P", 1);
	usartWriteString(str);
	positionReady = false;
}

boolean positionDataReady(position_type * position)
{
	if (positionReady)
	{
		*position = pos;
		return true;
	}
	return false;
}

void getHeading()
{
	char str[4];
	addChecksum(str, "H", 1);
	usartWriteString(str);
	headingReady = false;
}

boolean headingDataReady(unsigned int * heading_p)
{
	if (headingReady)
	{
		*heading_p = heading;
		return true;
	}
	return false;
}

static char generateChecksum(char a[], int n)
{
	int sum = 0;
	for (int i = 0; i < n; i++)
	{
		sum += (int)a[i];
	}
	return (char)(sum % 90 + (int)(' '));
}

boolean obstacleHit()
{
	return hasHitObstacle;
}

static void addChecksum(char * dest, char a[], int n)
{
	int sum = 0;
	for (int i = 0; i < n; i++)
	{
		dest[i] = a[i];
		sum += (int)a[i];
	}
	dest[n] = (char)(sum % 90 + (int)(' '));
	dest[n+1] = '\r';
	dest[n+2] = '\0';
}

static boolean checksumOK(char a[], int n)
{
	return a[n] == generateChecksum(a, n);
}


